/**
 * @file  AE3Modbus.cpp
 * @brief Implementierung der Klasse CAE3Modbus.
 */
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "AE3Modbus.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

// Umsetzung von CCom - auf CChannel - Fehlercodes 
const int g_iErrorMap[] = {
ERROR_NO,             /* ERR_NO */ 
ERROR_COM_INIT,       /* ERR_COM_OPEN */
ERROR_NO_ANSWER,      /* ERR_NO_COM */
ERROR_AT_SENDING,     /* ERR_WRITE */
ERROR_AT_RECIEVING,   /* ERR_READ */
ERROR_OPEN,           /* ERR_FILE_OPEN */
ERROR_CRC,            /* ERR_CRC */
ERROR_SENDER,         /* ERR_ADD_SE */
ERROR_COM,            /* ERR_BEFEHL_S4 */
ERROR_COM,            /* ? */
ERROR_FORMAT,         /* ERR_FORMAT */
ERROR_COM,            /* ? */
ERROR_BREAK,          /* ERR_ABBRUCH */
ERROR_COM,            /* ERR_UNBEKANNT */
ERROR_COM,            /* ERR_DAT_WRITE */
ERROR_COM,            /* ERR_WRITE_CNT */
ERROR_COM,            /* ERR_BEDIEN_STOP */
ERROR_COM             /* ERR_TYP_UNBEKANNT */
};

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

CAE3Modbus::CAE3Modbus(unsigned char ucChannelNum) : CChannel(ucChannelNum)
{
	m_ucBusNum = 0;

	// CallBack nicht definiert
	CBFunction = NULL;
}

CAE3Modbus::~CAE3Modbus()
{

}

//////////////////////////////////////////////////////////////////////
// Members
//////////////////////////////////////////////////////////////////////

void CAE3Modbus::ULongToBinMSBFirst(unsigned char *Buf,unsigned long Val)
{
	Buf[3] = (unsigned char)(Val&0xFF);
	Val >>= 8;
	Buf[2] = (unsigned char)(Val&0xFF);
	Val >>= 8;
	Buf[1] = (unsigned char)(Val&0xFF);
	Val >>= 8;
	Buf[0] = (unsigned char)(Val&0xFF);
}

void CAE3Modbus::UIntToBinMSBFirst(unsigned char *Buf,unsigned int Val)
{
	Buf[1] = (unsigned char)(Val&0xFF);
	Val >>= 8;
	Buf[0] = (unsigned char)(Val&0xFF);
}

int CAE3Modbus::BinToUIntMSBFirst(unsigned char *Buf)
{
	return ((int)Buf[0]<<8)|(int)Buf[1];

}

// Sende und Empfang von Modbus - Paket
int CAE3Modbus::Command(unsigned char *pucTxBuf, int iTxCount, 
						unsigned char *pucRxBuf, int *piRxCount, 
						int *piError)
{
	// Paketlaenge
	unsigned int uiPaketLength;
	// Anzahl geschriebener oder gelesender Zeichen
	ULONG ulReadWriteCount = 0;


	// Nicht aktuelle Busnummer
	if (pucTxBuf[0] != m_ucBusNum) { 

		ComLog("Command: Gesetzte Busnummer ist nicht die des zu sendenen Paketes.\n");	

		*piError = ERROR_BUS_NUM;
		return FALSE;
	}

	// Anfrage auf Kommunikationsklasse umlenken
	if (m_ChannelType == Class) {
	
		unsigned char ucBuf[V_MAX_PAKET];
		int iLength = iTxCount;

#ifdef _DEBUG
		memset(ucBuf, 0, V_MAX_PAKET);
		for (int i=0; i < V_MAX_PAKET; i++) {
			
			m_Com.SBuf[i] = 0;
			m_Com.EBuf[i] = 0;
		}
#endif
		memcpy(ucBuf, pucTxBuf, iLength);

		if (m_Com.OldMB(m_ucBusNum, ucBuf, &iLength, piError)) {

			memcpy(pucRxBuf, ucBuf, iLength);
			*piRxCount = iLength;

			return TRUE;
		}

		TRACE("ERROR_10: %i\n", *piError);

		// Fehlercode umsetzten
		*piError = g_iErrorMap[*piError];

		return FALSE;
	}

//@@: Callback mit Parameter andenken.
	if (!ExeCallBack()) {
	
		*piError = ERROR_BREAK;
		return FALSE;
	}

#ifdef _DEBUG
#define LOG	TRACE
#else
#define LOG ComLog
#endif

	// Befehl losschicken
	if (!Write((void*)pucTxBuf, iTxCount, &ulReadWriteCount)){
		
		LOG("Command: Paket konnte nicht versendet werden.\n");	
		
		*piError = ERROR_AT_SENDING;
		return FALSE;
	}

	if (ulReadWriteCount != (unsigned)iTxCount) {
	
		LOG("Command: Paket wurde nicht vollstaendig versendet.(%i) von (%i)\n", ulReadWriteCount, iTxCount);	

		// zu wendig gesendet	
		*piError = ERROR_AT_SENDING;
		return FALSE;
	}

	// Die ersten 4 Bytes der Antwort lesen
	if (!Read((void*)pucRxBuf, 4, &ulReadWriteCount)) {
			
		LOG("Command: Der Paketkopf konnte nicht empfangen werden.\n");	

		// AW1 kann nicht gelesen werden
		*piError = ERROR_NO_ANSWER;
		return FALSE; 
	}

	// Empfang pruefen
	if (ulReadWriteCount != 4) {
		
		LOG("Command: Es wurde ein zu kurzer Paketkopf empfangen.(%i) von (%i)\n", ulReadWriteCount, 4);	

		// Schnittstelle fuer naechste Anfrage bereinigen
		ClearCom();

		// AW1 zu kurz
		if(ulReadWriteCount == 0) {
			
			*piError = ERROR_NO_ANSWER;
		} else {
			
			*piError = ERROR_AT_RECIEVING;
		}

		return FALSE;
	}

	if (pucTxBuf[0] != pucRxBuf[0]) {

		LOG("Command: Busnummer von Sender und Empfaenger ungleich Tx(%i): Rx(%i)\n", pucTxBuf[0], pucRxBuf[0]);	

		// Schnittstelle fuer naechste Anfrage bereinigen
		ClearCom();

		// Absender falsch	
		*piError = ERROR_SENDER;
		return FALSE; 
	}

	// Die Laenge vom Paket lesen
	uiPaketLength = BinToUIntMSBFirst(&pucRxBuf[1]);

	// Rest vom Paket lesen
	if(!Read((void*)&pucRxBuf[4], uiPaketLength - 4, &ulReadWriteCount)){

		LOG("Command: Rest des Paketes lesens ist fehlgeschlagen.(%i) von (%i)\n", uiPaketLength, ulReadWriteCount+4);	

		// AW2 kann nicht gelesen werden
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	if (ulReadWriteCount != (uiPaketLength - 4)) {
		
		LOG("Command: Die gelesene Paketlaenge stimmt nicht mit der angegebenen ueberein.(%i) von (%i)\n", ulReadWriteCount, uiPaketLength);	

		// Schnittstelle fuer naechste Anfrage bereinigen
		ClearCom();

		// AW2 zu kurz
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	// Paketgroesse zurueckgeben
	*piRxCount = uiPaketLength;

	// Kommando erfolgreich ausgefuehrt
	return TRUE;

}

// CRC berechnen
void CAE3Modbus::MkMBPaket(unsigned char *buf, int cnt)	
{
	int i, j;
	unsigned short int crc;
  
	crc = 0xFFFF;
    for (i=0; i < cnt; i++) {

        crc = crc ^ buf[i];

        for (j=0; j < 8; j++) {
            if (crc & 0x1) {
				crc >>= 1; 
				crc = crc^0xA001;
			} else {
				crc >>= 1;
			}
        }
    }
    buf[cnt    ] = (unsigned char)(crc & 0xFF);
    buf[cnt + 1] = (unsigned char)(crc >> 8);
}

int CAE3Modbus::CheckMBPaket(unsigned char *buf, int cnt)
{
    unsigned short int crc;
    int i,j;

    crc = 0xFFFF;
    for (i=0; i < (cnt - 2); i++) {

        crc = crc ^ buf[i];

        for (j=0; j < 8; j++) {

            if (crc & 0x1) {
				crc >>= 1; 
				crc = crc^0xA001;
			} else {
				crc >>= 1;
			}
        }
    }
    
	if ( (buf[cnt - 2] != (unsigned char)(crc & 0xFF)) || 
		 (buf[cnt - 1] != (unsigned char)(crc >> 8)) ) { 
		
		return FALSE; 
	}

    return TRUE;
}

// Schreiben in AEIII+ (High-Level)
// Add -> Byte Adresse
// Cnt -> Anzahl Zeichen (max. 1MByte)
// Buf -> Buffer fuer empfangene Zeichen
int CAE3Modbus::WriteMem(unsigned long Add, int Cnt, unsigned char *Buf, int *piError, int MaxB)
{
	int i, j, ac, End = FALSE, ok;

	// Fehler loeschen
	*piError = ERROR_NO;

	// Bis zu (Gesamtpaket[ 0xFF ] - Overhead[ MAX_MB_PAKET_HEADER ]) Byte Daten in Geraet schreiben
	if (MaxB > MAX_MB_PAKET_DATA) {
	
		*piError = ERROR_RANGE;
		return FALSE;
	}
	// Bei PZE-Master IV mit maximaler Paketgroesse arbeiten
	if (GetChannelType() == Class && MaxB != 0) {

		MaxB = MAX_MB_PAKET_DATA;
	} else if (MaxB == 0) {
		// Wenn EEPROM geschrieben wird
		MaxB = 8;
	}

	// Daten in Pakete zerlegen
	for (i=0; !End; i += MaxB) {

		// Aufteilen
		if ((Cnt - i) < MaxB) {

			ac = Cnt - i;
		} else {

			ac = MaxB;
		}

		// Noch zu versendende Daten
		if ( ac > 0 ) {

			ok = FALSE;
			
			// Senden in mehreren Versuchen
			for (j=0; j < 4 && !ok; j++) {

				if (WriteXByte(Add + i, ac, &Buf[i], piError)) {

					ok = TRUE;
				} else {

					// Abbruch durch CallBack!
					if (*piError == ERROR_BREAK) {
						
						return FALSE;
					}

					Sleep(50);
				}
			}

			// Senden Fehlgeschlagen?
			if (!ok) {

				// Letzten Fehler zurueckgeben
				return FALSE;
			}

		} else {

			// Alle versendet
			End = TRUE;
		}
	}

	return TRUE;
}

// X Byte in AEIII+ schreiben (Low-Level)
// Add -> Byte Adresse
// Cnt -> Anzahl Zeichen (max. 242 Byte)
// Buf -> Buffer fuer empfangene Zeichen
int CAE3Modbus::WriteXByte (unsigned long Add, int Cnt, unsigned char *Buf, int *piError)
{
	int ret, i, h;
	unsigned char WBuf[MAX_WBUF];
	unsigned char RBuf[MAX_RBUF];

	if (Cnt > MAX_MB_PAKET_DATA) {
	
		*piError = ERROR_RANGE;
		return FALSE;
	}

	// Modbus-Befehl aufbauen
	WBuf[0] = m_ucBusNum;
	h = Cnt + 10;
	WBuf[1] = (unsigned char)(h >> 8);
	WBuf[2] = (unsigned char)(h & 0xFF);	// Laenge des Paketes
	WBuf[3] = MB_WRITE;
	ULongToBinMSBFirst(&WBuf[4], Add);
	
	// Daten in Paket kopieren
	for (i=0; i < Cnt; i++) {

		WBuf[i + 8] = Buf[i];
	}
	MkMBPaket(WBuf, 8 + Cnt); 
	ret = -1;

	bool bCRCFail = false;

crc_fail:

	if (!Command(WBuf, 10 + Cnt, RBuf, &ret, piError)) {
		
		return FALSE;
	}

	// Mindestens 6 fuer Fehlerpaket
	if (ret <= 4) {
	
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	if (!CheckMBPaket(RBuf, ret)) {

		if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

			bCRCFail = true;
			goto crc_fail;
		}

		*piError = ERROR_CRC;
		return FALSE;
	}

	if (RBuf[3] == MB_WRITE && ret == 12) {

		return TRUE;

	} else if (RBuf[3] == (MB_WRITE | 0x80) && ret == 7 && RBuf[4] == 0x22) {

		Sleep(150);
		return TRUE;

	} 

	// Antwort vom Falschen Geraet oder nicht sinnvoll
	*piError = ERROR_SENDER;
	return FALSE;
}

// Lesen aus AEIII+ (High-Level) 
// Add -> Byte Adresse
// Cnt -> Anzahl Zeichen (max. 1MByte)
// Buf -> Buffer fuer empfangene Zeichen
int CAE3Modbus::ReadMem(unsigned long Add, int Cnt, unsigned char *Buf, int *piError)
{
	int i, j, ac, End = FALSE, ok, MaxB = 32;

	// Fehler loeschen
	*piError = ERROR_NO;

	// Bei PZE-Master IV mit maximaler Paketgroesse arbeiten
	if (GetChannelType() == Class) {

		MaxB = MAX_MB_PAKET_DATA;
	} 

	// Bis zu 32 Byte pro Paket aus Geraet lesen (anfordern).
	// Daten in Paket zerlegen
	for (i=0; !End; i += MaxB) {

		// Zerlegen
		if ((Cnt - i) < MaxB) {

			ac = Cnt - i;
		} else {

			ac = MaxB;
		}

		// Daten vorhanden?
		if (ac > 0) {

			ok = FALSE;

			// Versuchtes anfragen
			for (j=0; j < 4 && !ok; j++) {

				if (ReadXByte(Add + i, ac, &Buf[i], piError)) {
					
					ok = TRUE; 
				} else {
					
					// Abbruch durch CallBack!
					if (*piError == ERROR_BREAK) {
						
						return FALSE;
					}

					Sleep(50);
				}
			}
			
			// Fehlgeschlagen
			if (!ok) {

				// Letzten Fehler zurueckgeben
				return FALSE;
			}

		} else {

			End = TRUE;
		}
	}

	return TRUE;
}

// X Byte von AEIII+ lesen (Low-Level)
// Add -> Byte Adresse
// Cnt -> Anzahl Zeichen (max. 32 Byte)
// Buf -> Buffer fuer empfangene Zeichen
int CAE3Modbus::ReadXByte (unsigned long Add, int Cnt, unsigned char *Buf, int *piError)
{
	int ret, i;
	unsigned char WBuf[MAX_WBUF];
	unsigned char RBuf[MAX_RBUF];

	if (Cnt > MAX_MB_PAKET_DATA) {
		
		*piError = ERROR_RANGE;
		return FALSE;
	}

	// Modbus-Befehl aufbauen
	WBuf[0] = m_ucBusNum;
	WBuf[1] = 0;
	WBuf[2] = 0x0C;		// Laenge des Paketes
	WBuf[3] = MB_READ;
	ULongToBinMSBFirst(&WBuf[4], Add);
	UIntToBinMSBFirst(&WBuf[8], Cnt);
	MkMBPaket(WBuf, 10);
	ret = -1;

	bool bCRCFail = false;

crc_fail:

	if (!Command(WBuf, 12, RBuf, &ret, piError)) {

		return FALSE;
	}

	if (ret <= 4) {
	
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	// Paket pruefen
	if (!CheckMBPaket(RBuf, ret)) {
	
		if (!bCRCFail) { // nach CRC - Fehler nochmals anfordern

			bCRCFail = true;
			goto crc_fail;
		}

		*piError = ERROR_CRC;
		return FALSE;
	}

	if (RBuf[3] == (MB_READ | 0x80) && ret == 7 && RBuf[4] == 0x22) {

		Sleep(150);
	}

	if (BinToUIntMSBFirst(&RBuf[1]) != (Cnt + 10)) {
	
		*piError = ERROR_AT_RECIEVING;
		return FALSE;
	}

	// auf Fehler pruefen
	if (RBuf[3] != MB_READ || RBuf[0] != m_ucBusNum) {

		TRACE("Fehler Befehl\n");
		*piError = ERROR_SENDER;
		return FALSE;
	}

	// Adresse pruefen
	for (i=0; i < 4; i++) {

		if (RBuf[i+4] != WBuf[i+4]) {

			*piError = ERROR_AT_RECIEVING;
			return FALSE;
		}
	}

	// Daten kopieren
	for (i=0; i < Cnt; i++) {

		Buf[i] = RBuf[i + 8];
	}

	return TRUE;
}

int CAE3Modbus::ChangeBusNum(unsigned char ucNewBusNum)
{
	if ((ucNewBusNum > 0 && ucNewBusNum < 32) || 
		ucNewBusNum == MB_ADD_OFFLINE) {
	
		m_ucBusNum = ucNewBusNum;
		return TRUE;
	} else {
	
		m_ucBusNum = 0;
		return FALSE;
	} 

	return FALSE;
}

unsigned char CAE3Modbus::GetBusNum() 
{
	return m_ucBusNum;
}

void CAE3Modbus::SetCBFunction(int (*pFunction)()) 
{
	CBFunction = pFunction;
	// Zeiger auf CallBack an Kommunikationsklasse weiterreichen.
	m_Com.SetCallBack(pFunction);
}

int CAE3Modbus::ExeCallBack()
{
	if (CBFunction) {

		return CBFunction();

	} else {

		return TRUE;
	}
}

